import typing


def printable_set(some_set: typing.Set[typing.Any]) -> str:
    """Get a printable, deterministic string version of a set."""
    return ", ".join([f"'{set_item}'" for set_item in sorted(list(some_set))])


def printable_set_coloured(some_set: typing.Set[typing.Any], colour: str) -> str:
    """`printable_set` but elements are surrounded with colour mark-up.

    Parameters:
        some_set: a set of anything
        colour: a `rich` Console supported colour
    """
    return ", ".join(
        [f"'[{colour}]{set_item}[/{colour}]'" for set_item in sorted(list(some_set))]
    )



import typing

import pytest

from runem.informative_dict import InformativeDict, ReadOnlyInformativeDict

# Define type variables for key and value to be used in the custom dictionary
K = typing.TypeVar("K")
V = typing.TypeVar("V")


@pytest.fixture(name="sample_dict")
def sample_dict_fixture() -> InformativeDict[str, int]:
    """Test fixture for creating an instance of InformativeDict."""
    return InformativeDict({"one": 1, "two": 2, "three": 3})


def test_getitem_existing_key(sample_dict: InformativeDict[str, int]) -> None:
    assert sample_dict["one"] == 1, (
        "Should retrieve the correct value for an existing key"
    )


def test_getitem_non_existent_key(sample_dict: InformativeDict[str, int]) -> None:
    with pytest.raises(KeyError) as exc_info:
        _ = sample_dict["four"]
    assert "Key 'four' not found. Available keys: one, two, three" in str(
        exc_info.value
    ), "Should raise KeyError with a message listing available keys"


def test_iteration(sample_dict: InformativeDict[str, int]) -> None:
    keys = list(sample_dict)
    assert keys == [
        "one",
        "two",
        "three",
    ], "Iteration over dictionary should yield all keys"


def test_initialization_with_items() -> None:
    dict_init = InformativeDict({"a": 1, "b": 2})
    assert dict_init["a"] == 1 and dict_init["b"] == 2, (
        "Should correctly initialize with given items"
    )


def test_read_only_get_existing_key() -> None:
    """Tests accessing valid ReadOnlyInformativeDict key returns correct value."""
    ro_dict = ReadOnlyInformativeDict[str, str]({"existing_key": "value"})
    assert ro_dict["existing_key"] == "value", (
        "Failed to retrieve the value for an existing key."
    )


def test_read_only_get_non_existent_key() -> None:
    """Tests attempted access of non-existent ReadOnlyInformativeDict key raises.

    It should raise a KeyError, and the error message includes the missing key and lists
    available keys.
    """
    ro_dict = ReadOnlyInformativeDict[str, str]({"existing_key": "value"})
    with pytest.raises(KeyError) as exc_info:
        _ = ro_dict["non_existent_key"]
    assert "non_existent_key" in str(exc_info.value), (
        "The exception message does not mention the missing key."
    )
    assert "existing_key" in str(exc_info.value), (
        "The exception message does not list available keys."
    )


# Define the operation type with concrete types for this test
OperationType = typing.Callable[..., None]


@pytest.mark.parametrize(
    "operation,args",
    [
        (lambda d, k, v: d.__setitem__(k, v), ("new_key", "new_value")),
        (lambda d, k, v: d.__delitem__(k), ("existing_key", None)),
        (lambda d, k, v: d.pop(k), ("existing_key", None)),
        (lambda d, k, v: d.popitem(), (None, None)),
        (lambda d, k, v: d.clear(), (None, None)),
        (lambda d, k, v: d.update({"another_key": "another_value"}), (None, None)),
    ],
)
def test_read_only_modification_operations_raise_error(
    operation: OperationType,
    args: typing.Tuple[typing.Optional[str], typing.Optional[str]],
) -> None:
    """Test that all modification operations on ReadOnlyInformativeDict (like setting an
    item, deleting an item, popping an item, clearing, and updating the dictionary)
    raise NotImplementedError.

    The test uses different operations and arguments to ensure comprehensive coverage.
    """
    ro_dict = ReadOnlyInformativeDict({"existing_key": "value"})
    with pytest.raises(NotImplementedError):
        operation(ro_dict, *args)



import typing

# meta-data types
JobName = str
JobTag = str
JobNames = typing.Set[JobName]
JobPhases = typing.Set[str]
JobTags = typing.Set[JobTag]
PhaseName = str
OrderedPhases = typing.Tuple[PhaseName, ...]
FilePathSerialise = str
FilePathList = typing.List[FilePathSerialise]



import typing
from collections import defaultdict

from runem.config_metadata import ConfigMetadata
from runem.job import Job
from runem.log import log
from runem.types.common import JobNames, JobPhases, JobTags, PhaseName
from runem.types.runem_config import JobConfig, PhaseGroupedJobs
from runem.utils import printable_set


def _should_filter_out_by_tags(
    job: JobConfig,
    tags: JobTags,
    tags_to_avoid: JobTags,
    verbose: bool,
) -> bool:
    job_tags: typing.Optional[JobTags] = Job.get_job_tags(job)
    opted_into_tag_filtering: bool = job_tags is not None
    if not opted_into_tag_filtering:
        # TODO: should we also consider 'empty' tags and being opted out of
        #       tag-filtering?
        return False

    # the config for the command has tags, use them
    assert job_tags is not None  # for mypy
    matching_tags = job_tags.intersection(tags)
    if not matching_tags:
        if verbose:
            log(
                (
                    f"not running job '{Job.get_job_name(job)}' because it doesn't have "
                    f"any of the following tags: {printable_set(tags)}"
                )
            )
        return True  # filter-out this job

    has_tags_to_avoid = job_tags.intersection(tags_to_avoid)
    if has_tags_to_avoid:
        if verbose:
            log(
                (
                    f"not running job '{Job.get_job_name(job)}' because it contains the "
                    f"following tags: {printable_set(has_tags_to_avoid)}"
                )
            )
        return True  # filter-out this job

    # no filter criteria met, filter-in the job i.e. run the job
    return False


def _get_jobs_matching(
    phase: PhaseName,
    job_names: JobNames,
    tags: JobTags,
    tags_to_avoid: JobTags,
    jobs: PhaseGroupedJobs,
    filtered_jobs: PhaseGroupedJobs,
    verbose: bool,
) -> None:
    """Via filtered_jobs, filters 'jobs' that match the given phase and and tags.

    Warns if the job-name isn't found in list of valid job-names.
    """
    phase_jobs: typing.List[JobConfig] = jobs[phase]

    job: JobConfig
    for job in phase_jobs:
        if _should_filter_out_by_tags(job, tags, tags_to_avoid, verbose):
            continue

        job_name: str = Job.get_job_name(job)
        if job_name not in job_names:
            # test test_get_jobs_matching_when_job_not_in_valid_job_names should
            # cover the follow in Ci but does not for some reason I don't have
            # time to look in to. /FH
            if verbose:  # pragma: FIXME: add code coverage
                log(
                    (
                        f"not running job '{job_name}' because it isn't in the "
                        f"list of job names. See --jobs and --not-jobs"
                    )
                )
            continue

        filtered_jobs[phase].append(job)


def filter_jobs(  # noqa: C901
    config_metadata: ConfigMetadata,
) -> PhaseGroupedJobs:
    """Filters the jobs to match requested tags."""
    jobs_to_run: JobNames = config_metadata.jobs_to_run
    phases_to_run: JobPhases = config_metadata.phases_to_run
    tags_to_run: JobTags = config_metadata.tags_to_run
    tags_to_avoid: JobTags = config_metadata.tags_to_avoid
    jobs: PhaseGroupedJobs = config_metadata.jobs
    verbose: bool = config_metadata.args.verbose
    if verbose:
        if tags_to_run:
            log(
                f"filtering for tags {printable_set(tags_to_run)}",
                prefix=True,
                end="",
            )
        if tags_to_avoid:
            if tags_to_run:
                log(", ", prefix=False, end="")
            else:
                log(prefix=True, end="")
            log(
                f"excluding jobs with tags {printable_set(tags_to_avoid)}",
                prefix=False,
                end="",
            )
        if tags_to_run or tags_to_avoid:
            log(prefix=False)
    filtered_jobs: PhaseGroupedJobs = defaultdict(list)
    for phase in config_metadata.phases:
        if phase not in phases_to_run:
            # test test_get_jobs_matching_when_job_not_in_valid_job_names should
            # cover the follow in Ci but does not for some reason I don't have
            # time to look in to. /FH
            if verbose:  # pragma: FIXME: add code coverage
                log(f"skipping phase '{phase}'")
            continue
        _get_jobs_matching(
            phase=phase,
            job_names=jobs_to_run,
            tags=tags_to_run,
            tags_to_avoid=tags_to_avoid,
            jobs=jobs,
            filtered_jobs=filtered_jobs,
            verbose=verbose,
        )
        if len(filtered_jobs[phase]) == 0:
            # test test_get_jobs_matching_when_job_not_in_valid_job_names should
            # cover the follow in Ci but does not for some reason I don't have
            # time to look in to. /FH
            if verbose:  # pragma: FIXME: add code coverage
                log(f"No jobs for phase '{phase}' tags {printable_set(tags_to_run)}")
            continue

        if verbose:
            log((f"will run {len(filtered_jobs[phase])} jobs for phase '{phase}'"))
            job_names: JobNames = {
                Job.get_job_name(job) for job in filtered_jobs[phase]
            }
            log(f"\t{printable_set(job_names)}")

    return filtered_jobs



import typing

from runem.types.common import FilePathList, JobTags
from runem.types.filters import FilePathListLookup
from runem.types.runem_config import JobConfig, JobWhen


class NoJobName(ValueError):
    """The job-config does not contain a label and can't be coerced to crate one."""

    pass


class BadWhenConfigLocation(ValueError):
    """The job-config does not contain a label and can't be coerced to crate one."""

    pass


class Job:
    """A class with utility functions for jobs.

    Currently these are static but eventually will be members
    """

    @staticmethod
    def get_job_tags(job: JobConfig) -> typing.Optional[JobTags]:
        """Returns the tags for this job or None if the tags key is missing.

        In practice None means:
        - get all files
        - always run job

        TODO: make a non-static member function
        """
        if "tags" in job:
            raise BadWhenConfigLocation(
                "'tags' should be listed inside the 'when' config for jobs"
            )

        if "phase" in job:
            raise BadWhenConfigLocation(
                "'phase' should be listed inside the 'when' config for jobs"
            )

        if "when" not in job or "tags" not in job["when"]:
            # handle the special case where we have No tags
            return None

        # have valid tags, coerce them to be a set-type and return
        when: JobWhen = job.get("when", {})
        job_tags: JobTags = when["tags"]
        return set(job_tags)

    @staticmethod
    def get_job_files(
        file_lists: FilePathListLookup, job_tags: typing.Optional[JobTags]
    ) -> FilePathList:
        """Return the list of files for the job-associated tags.

        TODO: support no files-groups being defined i.e. switch to all files
              being used (maybe).

        FIXME: this is a bad design choice that will need revisiting. I guess
               the options are to:
            1. Have file-list-tags (ugh)
            2. Keep associating tags with file-sets
            3. Add additional filtering based on cwd (ugh)

            #1 is probably the winner but we'll ponder on it before deciding.

        TODO: make a non-static member function
        """
        # default to all file-tags
        tags_for_files: typing.Iterable[str] = file_lists.keys()
        use_default_tags: bool = job_tags is None
        if not use_default_tags:
            # use whatever tags are in the config, even if empty
            assert job_tags  # for mypy
            tags_for_files = job_tags
        file_list: FilePathList = []
        for tag in tags_for_files:
            if tag in file_lists:
                file_list.extend(file_lists[tag])
        return sorted(file_list)

    @staticmethod
    def get_job_name(job: JobConfig) -> str:
        """Returns a name to use for a given job config.

        TODO: make a non-static member function
        """
        # First try one of the following keys.
        valid_name_keys = ("label", "command")
        for candidate in valid_name_keys:
            name: typing.Optional[str] = job.get(candidate, None)  # type: ignore # NO_COMMIT
            if name:
                return name

        # The try the python-wrapper address
        try:
            return f"{job['addr']['file']}.{job['addr']['function']}"
        except KeyError:
            raise NoJobName()  # pylint: disable=raise-missing-from



import argparse
import os
import pathlib
import sys
import typing

from runem.config_metadata import ConfigMetadata
from runem.informative_dict import InformativeDict
from runem.log import error, log
from runem.runem_version import get_runem_version
from runem.types.common import JobNames
from runem.types.options import OptionsWritable
from runem.types.runem_config import OptionConfig
from runem.utils import printable_set


class HelpFormatterFixedWidth(argparse.HelpFormatter):
    """This works around test issues via constant width helo output.

    This ensures that we get more constant for help-text by fixing the width to
    something reasonable.
    """

    def __init__(self, prog: typing.Any) -> None:
        # Override the default width with a fixed width, for tests.
        super().__init__(
            prog,
            # Pretty wide so we do not get wrapping on directories
            # or process-count output.
            width=1000,
        )


def _get_argparse_help_formatter() -> typing.Any:
    """Returns a help-formatter for argparse.

    This is for tests only to fake terminals of a constant with when rendering help
    output.
    """
    # Check environment variable to see if we're in tests and need a fixed width
    # help output.
    use_fixed_width = os.getenv("RUNEM_FIXED_HELP_WIDTH", None)

    if use_fixed_width:
        # Use custom formatter with the width specified in the environment variable
        return lambda prog: HelpFormatterFixedWidth(  # pylint: disable=unnecessary-lambda
            prog
        )

    # Use default formatter
    return argparse.HelpFormatter


def error_on_log_logic(verbose: bool, silent: bool) -> None:
    """Simply errors if we get logical inconsistencies in the logging-logic."""
    if verbose and silent:
        log("cannot parse '--verbose' and '--silent'")
        # error exit
        sys.exit(1)


def parse_args(
    config_metadata: ConfigMetadata, argv: typing.List[str]
) -> ConfigMetadata:
    """Parses the args and defines the filter inputs.

    Generates args based upon the config, parsing the cli args and return the filters to
    be used when selecting jobs.

    Returns the parsed args, the jobs_names_to_run, job_phases_to_run, job_tags_to_run
    """
    parser = argparse.ArgumentParser(
        add_help=False,
        description="Runs the Lursight Lang test-suite",
        formatter_class=_get_argparse_help_formatter(),
    )
    parser.add_argument(
        "-H", "--help", action="help", help="show this help message and exit"
    )

    job_group = parser.add_argument_group("jobs")
    all_job_names: JobNames = set(name for name in config_metadata.all_job_names)
    job_group.add_argument(
        "--jobs",
        dest="jobs",
        nargs="+",
        default=sorted(list(all_job_names)),
        help=(
            "List of job-names to run the given jobs. Other filters will modify this list. "
            f"Defaults to {printable_set(all_job_names)}"
        ),
        required=False,
    )
    job_group.add_argument(
        "--not-jobs",
        dest="jobs_excluded",
        nargs="+",
        default=[],
        help=(
            "List of job-names to NOT run. Defaults to empty. "
            f"Available options are: {printable_set((all_job_names))}"
        ),
        required=False,
    )

    phase_group = parser.add_argument_group("phases")
    phase_group.add_argument(
        "--phases",
        dest="phases",
        nargs="+",
        default=config_metadata.all_job_phases,
        help=(
            "Run only the phases passed in, and can be used to "
            "change the phase order. Phases are run in the order given. "
            f"Defaults to {printable_set(config_metadata.all_job_phases)}. "
        ),
        required=False,
    )
    phase_group.add_argument(
        "--not-phases",
        dest="phases_excluded",
        nargs="+",
        default=[],
        help=(
            "List of phases to NOT run. "
            "This option does not change the phase run order. "
            f"Options are '{sorted(config_metadata.all_job_phases)}'. "
        ),
        required=False,
    )

    tag_group = parser.add_argument_group("tags")
    tag_group.add_argument(
        "--tags",
        dest="tags",
        nargs="+",
        default=config_metadata.all_job_tags,
        help=(
            # TODO: clarify the logic by which we add/remove jobs based on tags
            #       e.g. exclusive-in, union, x-or etc.
            "Only run jobs with the given tags. "
            f"Defaults to '{sorted(config_metadata.all_job_tags)}'."
        ),
        required=False,
    )
    tag_group.add_argument(
        "--not-tags",
        dest="tags_excluded",
        nargs="+",
        default=[],
        help=(
            "Removes one or more tags from the list of job tags to be run. "
            f"Options are '{sorted(config_metadata.all_job_tags)}'."
        ),
        required=False,
    )

    job_param_overrides_group = parser.add_argument_group(
        "job-param overrides",  # help="overrides default test params on all matching jobs"
    )
    _define_option_args(config_metadata, job_param_overrides_group)

    parser.add_argument(
        "--call-graphs",
        dest="generate_call_graphs",
        action=argparse.BooleanOptionalAction,
        default=False,
        required=False,
    )

    parser.add_argument(
        "-f",
        "--modified-files",
        dest="check_modified_files",
        help="only use files that have changed",
        action=argparse.BooleanOptionalAction,
        default=False,
        required=False,
    )

    parser.add_argument(
        "-h",
        "--git-head-files",
        dest="check_head_files",
        help="fast run of files",
        action=argparse.BooleanOptionalAction,
        default=False,
        required=False,
    )
    parser.add_argument(
        "--always-files",
        dest="always_files",
        help=(
            "list of paths/files to always check (overriding -f/-h), if the path "
            "matches the filter regex and if file-paths exist"
        ),
        nargs="+",
        default=None,
        required=False,
    )

    parser.add_argument(
        "--git-files-since-branch",
        dest="git_since_branch",
        help=(
            "Get the list of paths/files changed between a branch, e.g., since "
            "'origin/main'. Useful for checking files changed before pushing."
        ),
        default=None,  # Default to None if no branch is specified
        required=False,  # Not required, users may not want to specify a branch
        type=str,  # Accepts a string input representing the branch name
    )

    parser.add_argument(
        "--procs",
        "-j",
        # "-n",
        dest="procs",
        default=-1,
        help=(
            "the number of concurrent test jobs to run, -1 runs all test jobs at the same time "
            f"({os.cpu_count()} cores available)"
        ),
        required=False,
        type=int,
    )

    config_dir: pathlib.Path = _get_config_dir(config_metadata)
    parser.add_argument(
        "--root",
        dest="root_dir",
        default=config_dir,
        help=(
            "which dir to use as the base-dir for testing, "
            f"defaults to directory containing the config '{config_dir}'"
        ),
        required=False,
    )

    parser.add_argument(
        "--root-show",
        dest="show_root_path_and_exit",
        help="show the root-path of runem and exit",
        action=argparse.BooleanOptionalAction,
        default=False,
        required=False,
    )

    parser.add_argument(
        "--silent",
        "-s",
        dest="silent",
        action=argparse.BooleanOptionalAction,
        default=False,
        help=("Whether to show warning messages or not. "),
        required=False,
    )

    parser.add_argument(
        "--spinner",
        dest="show_spinner",
        action=argparse.BooleanOptionalAction,
        default=True,
        help=(
            "Whether to show the progress spinner or not. "
            "Helps reduce log-spam in ci/cd."
        ),
        required=False,
    )

    parser.add_argument(
        "--verbose",
        dest="verbose",
        help="runs runem in in verbose mode, and streams jobs stdout/stderr to console",
        action=argparse.BooleanOptionalAction,
        default=False,
        required=False,
    )

    parser.add_argument(
        "--version",
        "-v",
        dest="show_version_and_exit",
        help="show the version of runem and exit",
        action=argparse.BooleanOptionalAction,
        default=False,
        required=False,
    )

    args = parser.parse_args(argv[1:])

    error_on_log_logic(args.verbose, args.silent)

    if args.show_root_path_and_exit:
        log(str(config_metadata.cfg_filepath.parent), prefix=False)
        # cleanly exit
        sys.exit(0)

    if args.show_version_and_exit:
        log(str(get_runem_version()), prefix=False)
        # cleanly exit
        sys.exit(0)

    options: OptionsWritable = initialise_options(config_metadata, args)

    if not _validate_filters(config_metadata, args):
        sys.exit(1)

    # apply the cli filters to produce the high-level requirements. These will be used
    # to further filter the jobs.
    jobs_to_run = set(args.jobs).difference(args.jobs_excluded)
    tags_to_run = set(args.tags).difference(args.tags_excluded)
    tags_to_avoid = set(args.tags_excluded)
    phases_to_run = set(args.phases).difference(args.phases_excluded)

    config_metadata.set_cli_data(
        args, jobs_to_run, phases_to_run, tags_to_run, tags_to_avoid, options
    )
    return config_metadata


def _get_config_dir(config_metadata: ConfigMetadata) -> pathlib.Path:
    """A function to get the path, that we can mock in tests."""
    return config_metadata.cfg_filepath.parent


def _validate_filters(
    config_metadata: ConfigMetadata,
    args: argparse.Namespace,
) -> bool:
    """Validates the command line filters given.

    returns True of success and False on failure
    """
    # validate the job-names passed in
    for name, name_list in (("--jobs", args.jobs), ("--not-jobs", args.jobs_excluded)):
        for job_name in name_list:
            if job_name not in config_metadata.all_job_names:
                error(
                    (
                        f"invalid job-name '{job_name}' for {name}, "
                        f"choose from one of {printable_set(config_metadata.all_job_names)}"
                    )
                )
                return False

    # validate the tags passed in
    for name, tag_list in (("--tags", args.tags), ("--not-tags", args.tags_excluded)):
        for tag in tag_list:
            if tag not in config_metadata.all_job_tags:
                error(
                    (
                        f"invalid tag '{tag}' for {name}, "
                        f"choose from one of {printable_set(config_metadata.all_job_tags)}"
                    )
                )
                return False

    # validate the phases passed in
    for name, phase_list in (
        ("--phases", args.phases),
        ("--not-phases", args.phases_excluded),
    ):
        for phase in phase_list:
            if phase not in config_metadata.all_job_phases:
                error(
                    (
                        f"invalid phase '{phase}' for {name}, "
                        f"choose from one of {printable_set(config_metadata.all_job_phases)}"
                    )
                )
                return False
    return True


def initialise_options(
    config_metadata: ConfigMetadata,
    args: argparse.Namespace,
) -> OptionsWritable:
    """Initialises and returns the set of options to use for this run.

    Returns the options dictionary
    """
    options: OptionsWritable = InformativeDict(
        {option["name"]: option["default"] for option in config_metadata.options_config}
    )
    if config_metadata.options_config and args.overrides_on:  # pragma: no branch
        for option_name in args.overrides_on:  # pragma: no branch
            options[option_name] = True
    if config_metadata.options_config and args.overrides_off:  # pragma: no branch
        for option_name in args.overrides_off:
            options[option_name] = False
    return options


def _define_option_args(
    config_metadata: ConfigMetadata, job_param_overrides_group: argparse._ArgumentGroup
) -> None:
    option: OptionConfig
    for option in config_metadata.options_config:
        switch_name = option["name"].replace("_", "-").replace(" ", "-")

        aliases: typing.List[str] = []
        aliases_no: typing.List[str] = []
        if "aliases" in option and option["aliases"]:
            aliases = [
                _alias_to_switch(switch_name_alias)
                for switch_name_alias in option["aliases"]
            ]
            aliases_no = [
                _alias_to_switch(switch_name_alias, negatise=True)
                for switch_name_alias in option["aliases"]
            ]
        if "alias" in option and option["alias"]:
            aliases.append(_alias_to_switch(option["alias"]))
            aliases_no.append(_alias_to_switch(option["alias"], negatise=True))

        desc: typing.Optional[str] = None
        desc_for_off: typing.Optional[str] = None
        if "desc" in option:
            desc = option["desc"]
            desc_for_off = f"turn off {desc}"

        job_param_overrides_group.add_argument(
            f"--{switch_name}",
            *aliases,
            dest="overrides_on",
            action="append_const",
            const=option["name"],
            help=desc,
            required=False,
        )
        job_param_overrides_group.add_argument(
            f"--no-{switch_name}",
            *aliases_no,
            dest="overrides_off",
            action="append_const",
            const=option["name"],
            help=desc_for_off,
            required=False,
        )


def _alias_to_switch(switch_name_alias: str, negatise: bool = False) -> str:
    """Util function to generate a alias switch for argparse."""
    single_letter_variant = not negatise and len(switch_name_alias) == 1
    if single_letter_variant:
        return f"-{switch_name_alias}"
    if negatise:
        return f"--no-{switch_name_alias}"
    return f"--{switch_name_alias}"



import typing
from collections import defaultdict

from typing_extensions import Unpack

from runem.config_metadata import ConfigMetadata
from runem.job import Job
from runem.job_execute import job_execute
from runem.log import log
from runem.types.filters import FilePathListLookup
from runem.types.hooks import HookName
from runem.types.runem_config import HookConfig, Hooks, HooksStore, JobConfig
from runem.types.types_jobs import HookSpecificKwargs


class HookManager:
    hooks_store: HooksStore

    def __init__(self, hooks: Hooks, verbose: bool) -> None:
        self.hooks_store: HooksStore = defaultdict(list)
        self.initialise_hooks(hooks, verbose)

    @staticmethod
    def is_valid_hook_name(hook_name: typing.Union[HookName, str]) -> bool:
        """Returns True/False depending on hook-name validity."""
        if isinstance(hook_name, str):
            try:
                HookName(hook_name)  # lookup by value
                return True
            except ValueError:
                return False
        # the type is a HookName
        if not isinstance(hook_name, HookName):
            return False
        return True

    def register_hook(
        self, hook_name: HookName, hook_config: HookConfig, verbose: bool
    ) -> None:
        """Registers a hook_config to a specific hook-type."""
        if not self.is_valid_hook_name(hook_name):
            raise ValueError(f"Hook {hook_name} does not exist.")
        self.hooks_store[hook_name].append(hook_config)
        if verbose:
            log(
                f"hooks: registered hook for '{hook_name}', "
                f"have {len(self.hooks_store[hook_name])}: "
                f"{Job.get_job_name(hook_config)}"  # type: ignore[arg-type]
            )

    def deregister_hook(
        self, hook_name: HookName, hook_config: HookConfig, verbose: bool
    ) -> None:
        """Deregisters a hook_config from a specific hook-type."""
        if not (
            hook_name in self.hooks_store and hook_config in self.hooks_store[hook_name]
        ):
            raise ValueError(f"Function not found in hook {hook_name}.")
        self.hooks_store[hook_name].remove(hook_config)
        if verbose:
            log(
                f"hooks: deregistered hooks for '{hook_name}', "
                f"have {len(self.hooks_store[hook_name])}"
            )

    def invoke_hooks(
        self,
        hook_name: HookName,
        config_metadata: ConfigMetadata,
        **kwargs: Unpack[HookSpecificKwargs],
    ) -> None:
        """Invokes all functions registered to a specific hook."""
        hooks: typing.List[HookConfig] = self.hooks_store.get(hook_name, [])
        if config_metadata.args.verbose:
            log(f"hooks: invoking {len(hooks)} hooks for '{hook_name}'")

        hook_config: HookConfig
        for hook_config in hooks:
            job_config: JobConfig = {
                "label": str(hook_name),
                "ctx": None,
                "when": {"phase": str(hook_name), "tags": {str(hook_name)}},
            }
            if "addr" in hook_config:
                job_config["addr"] = hook_config["addr"]
            if "command" in hook_config:
                job_config["command"] = hook_config["command"]
            file_lists: FilePathListLookup = defaultdict(list)
            file_lists[str(hook_name)] = [__file__]
            job_execute(
                job_config,
                running_jobs={},
                completed_jobs={},
                config_metadata=config_metadata,
                file_lists=file_lists,
                **kwargs,
            )

        if config_metadata.args.verbose:
            log(f"hooks: done invoking '{hook_name}'")

    def initialise_hooks(self, hooks: Hooks, verbose: bool) -> None:
        """Initialised the hook with the configured data."""
        if verbose:
            num_hooks: int = sum(len(hooks_list) for hooks_list in hooks.values())
            if num_hooks:
                log(f"hooks: initialising {num_hooks} hooks")
        for hook_name in hooks:
            hook: HookConfig
            if verbose:
                log(
                    f"hooks:\tinitialising {len(hooks[hook_name])} hooks for '{hook_name}'"
                )
            for hook in hooks[hook_name]:
                self.register_hook(hook_name, hook, verbose)



import io
import pathlib
from contextlib import redirect_stdout
from datetime import timedelta
from typing import Optional, Tuple, Type
from unittest.mock import Mock, patch

import pytest

import runem.run_command


class MockPopen:
    """Mock Popen object."""

    def __init__(self, returncode: int = 0, stdout: str = "test output") -> None:
        self.returncode: int = returncode
        self.stdout: io.StringIO = io.StringIO(stdout)

    def communicate(self) -> Tuple[str, bytes]:  # pragma: no cover
        """Mock the communicate method if you use it."""
        # Assuming the stdout StringIO object's content should be returned as str
        return self.stdout.getvalue(), b""

    def __enter__(self) -> "MockPopen":
        """Mimic the behaviour of the context manager."""
        return self

    def __exit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_val: Optional[BaseException],
        exc_tb: Optional[Type[BaseException]],
    ) -> None:
        """Do nothing in the context on exit."""
        pass

    def wait(self) -> int:
        """Mimic wait() if used in your implementation."""
        return self.returncode


@pytest.mark.parametrize(
    "stdout, expected",
    [
        ("Line1\nLine2", "Prefix: Line1\nPrefix: Line2"),
        ("SingleLine", "Prefix: SingleLine"),
        ("Line1\nLine2\n", "Prefix: Line1\nPrefix: Line2\nPrefix: "),
        ("Prefix:\n", "Prefix: Prefix:\nPrefix: "),
        ("", "Prefix: "),
        ("Line1\n\nLine3", "Prefix: Line1\nPrefix: \nPrefix: Line3"),
        ("Line1\nLine2\n\n", "Prefix: Line1\nPrefix: Line2\nPrefix: \nPrefix: "),
    ],
    ids=[
        "multiple_lines",
        "single_line",
        "multiple_lines_with_trailing_newline",
        "single_newline",
        "empty_string",
        "lines_with_empty_line_in_between",
        "multiple_lines_with_multiple_trailing_newlines",
    ],
)
def test_parse_stdout(stdout: str, expected: str) -> None:
    """Test various scenarios for the parse_stdout function.

    NOTE: we assume non-bytes strings here.
    """
    assert expected == runem.run_command.parse_stdout(stdout, "Prefix: ")


@patch("runem.run_command.Popen", autospec=True, return_value=MockPopen())
def test_run_command_basic_call(mock_popen: Mock) -> None:
    """Test normal operation of the run_command.

    That is, that we can run a successful command and set the run-context for it
    """
    record_times_called: bool = False

    def dummy_record_sub_job_time(label: str, timing: timedelta) -> None:
        nonlocal record_times_called
        record_times_called = True

    # capture any prints the run_command() does, should be none in verbose=False mode
    with io.StringIO() as buf, redirect_stdout(buf):
        output = runem.run_command.run_command(
            cmd=["ls"],
            label="test command",
            verbose=False,
            record_sub_job_time=dummy_record_sub_job_time,
        )
        run_command_stdout = buf.getvalue()
    assert output == "test output"
    assert "" == run_command_stdout, "expected empty output when verbosity is off"
    mock_popen.assert_called_once()
    assert len(mock_popen.call_args) == 2
    assert mock_popen.call_args[0] == (["ls"],)
    call_ctx = mock_popen.call_args[1]
    env = call_ctx["env"]
    assert len(env.keys()) > 0, "expected the calling env to be passed to the command"
    assert record_times_called


@patch("runem.run_command.Popen", autospec=True, return_value=MockPopen())
def test_run_command_basic_call_verbose(mock_popen: Mock) -> None:
    """Test that we get extra output when the verbose flag is set."""
    # capture any prints the run_command() does, should be informative in verbose=True mode
    with io.StringIO() as buf, redirect_stdout(buf):
        raw_output = runem.run_command.run_command(
            cmd=["ls"], label="test command", verbose=True
        )
        run_command_stdout = buf.getvalue()
    assert raw_output == "test output"

    # check the log output hasn't changed. Update as needed.
    assert run_command_stdout == (
        "runem: running: start: test command: ls\n"
        "| test command: test output\n"
        "runem: running: done: test command: ls\n"
    )
    mock_popen.assert_called_once()


@patch(
    "runem.run_command.Popen",
    autospec=True,
    return_value=MockPopen(
        returncode=1,  # use an error-code of 1, FAIL
    ),
)
def test_run_command_basic_call_non_zero_exit_code(mock_popen: Mock) -> None:
    """Mimic non-zero exit code."""
    # capture any prints the run_command() does, should be informative in verbose=True mode
    with io.StringIO() as buf, redirect_stdout(buf):
        with pytest.raises(runem.run_command.RunCommandBadExitCode):
            runem.run_command.run_command(
                cmd=["ls"], label="test command", verbose=False
            )

        run_command_stdout = buf.getvalue()

    # check the log output hasn't changed. Update as needed.
    assert run_command_stdout == ""
    mock_popen.assert_called_once()


@patch(
    "runem.run_command.Popen",
    autospec=True,
    side_effect=ValueError,
)
def test_run_command_handles_throwing_command(mock_popen: Mock) -> None:
    """Mimic non-zero exit code."""
    # capture any prints the run_command() does, should be informative in verbose=True mode
    with io.StringIO() as buf, redirect_stdout(buf):
        with pytest.raises(runem.run_command.RunCommandUnhandledError):
            runem.run_command.run_command(
                cmd=["ls"], label="test command", verbose=False
            )

        run_command_stdout = buf.getvalue()

    # check the log output hasn't changed. Update as needed.
    assert run_command_stdout == ""
    mock_popen.assert_called_once()


@patch("runem.run_command.Popen", autospec=True, return_value=MockPopen(returncode=1))
def test_run_command_ignore_fails_skips_failures_for_non_zero_exit_code(
    mock_popen: Mock,
) -> None:
    """Mimic non-zero exit code, but ensure we do NOT raise when ignore_fails=True."""
    # capture any prints the run_command() does, should be informative in verbose=True mode
    with io.StringIO() as buf, redirect_stdout(buf):
        output = runem.run_command.run_command(
            cmd=["ls"],
            label="test command",
            verbose=False,
            ignore_fails=True,
        )
        assert output == "", (
            "expected empty output on failed run with 'ignore_fails=True'"
        )

        run_command_stdout = buf.getvalue()

    # check the log output hasn't changed. Update as needed.
    assert run_command_stdout == ""
    mock_popen.assert_called_once()


@patch(
    "runem.run_command.Popen",
    autospec=True,
    return_value=MockPopen(
        returncode=0,  # leave valid_exit_ids param at default of 0, no-error
    ),
)
def test_run_command_ignore_fails_skips_no_side_effects_on_success(
    mock_popen: Mock,
) -> None:
    """Mimic non-zero exit code, but ensure we do NOT raise when ignore_fails=True."""
    # capture any prints the run_command() does, should be informative in verbose=True mode
    with io.StringIO() as buf, redirect_stdout(buf):
        output = runem.run_command.run_command(
            cmd=["ls"],
            label="test command",
            verbose=False,
            ignore_fails=True,
        )
        assert output == "test output", (
            "expected empty output on failed run with 'ignore_fails=True'"
        )

        run_command_stdout = buf.getvalue()

    # check the log output hasn't changed. Update as needed.
    assert run_command_stdout == ""
    mock_popen.assert_called_once()


@patch(
    "runem.run_command.Popen",
    autospec=True,
    return_value=MockPopen(
        returncode=3,  # use 3, aka error code, but we will allow this later
    ),
)
def test_run_command_ignore_fails_skips_no_side_effects_on_success_with_valid_exit_ids(
    mock_popen: Mock,
) -> None:
    """Mimic non-zero exit code, but ensure we do NOT raise when ignore_fails=True."""
    # capture any prints the run_command() does, should be informative in verbose=True mode
    with io.StringIO() as buf, redirect_stdout(buf):
        output = runem.run_command.run_command(
            cmd=["ls"],
            label="test command",
            verbose=False,
            valid_exit_ids=(3,),  # matches patch value for 'returncode' above
            ignore_fails=True,
        )
        assert output == "test output", (
            "expected empty output on failed run with 'ignore_fails=True'"
        )

        run_command_stdout = buf.getvalue()

    # check the log output hasn't changed. Update as needed.
    assert run_command_stdout == ""
    mock_popen.assert_called_once()


@patch(
    "runem.run_command.Popen",
    autospec=True,
    return_value=MockPopen(
        returncode=3,  # set to 3 to mimic tools that return non-zero in aok modes
    ),
)
def test_run_command_basic_call_non_standard_exit_ok_code(mock_popen: Mock) -> None:
    """Tests the feature that handles non-standard exit codes."""
    # capture any prints the run_command() does, should be informative in verbose=True mode
    with io.StringIO() as buf, redirect_stdout(buf):
        output = runem.run_command.run_command(
            cmd=["ls"],
            label="test command",
            verbose=False,
            valid_exit_ids=(3,),  # matches the monkey-patch config about
        )
        run_command_stdout = buf.getvalue()
    assert output == "test output"

    # check the log output hasn't changed. Update as needed.
    assert run_command_stdout == ""
    mock_popen.assert_called_once()


@patch(
    "runem.run_command.Popen",
    autospec=True,
    return_value=MockPopen(
        returncode=3,  # set to 3 to mimic tools that return non-zero in aok modes
    ),
)
def test_run_command_basic_call_non_standard_exit_ok_code_verbose(
    mock_popen: Mock,
) -> None:
    """Tests we handle non-standard exit codes & log out extra relevant information."""
    # capture any prints the run_command() does, should be informative in verbose=True mode
    with io.StringIO() as buf, redirect_stdout(buf):
        output = runem.run_command.run_command(
            cmd=["ls"],
            label="test command",
            verbose=True,  # we expect the out to change with verbose AND valid_exit_ids
            valid_exit_ids=(3,),  # matches the monkey-patch config about
        )
        run_command_stdout = buf.getvalue()
    assert output == "test output"

    # check the log output hasn't changed. Update as needed.
    assert run_command_stdout == (
        "runem: running: start: test command: ls\n"
        "runem:  allowed return ids are: 3\n"
        "| test command: test output\n"
        "runem: running: done: test command: ls\n"
    )
    mock_popen.assert_called_once()


@patch("runem.run_command.Popen", autospec=True, return_value=MockPopen())
def test_run_command_with_env(mock_popen: Mock) -> None:
    """Tests that the env is passed to the subprocess."""
    # capture any prints the run_command() does, should be none in verbose=False mode
    with io.StringIO() as buf, redirect_stdout(buf):
        output = runem.run_command.run_command(
            cmd=["ls"],
            label="test command",
            verbose=False,
            env_overrides={"TEST_ENV_1": "1", "TEST_ENV_2": "2"},
        )
        run_command_stdout = buf.getvalue()
    assert output == "test output"
    assert "" == run_command_stdout, "expected empty output when verbosity is off"
    assert len(mock_popen.call_args) == 2
    assert mock_popen.call_args[0] == (["ls"],)
    call_ctx = mock_popen.call_args[1]
    env = call_ctx["env"]
    assert "TEST_ENV_1" in env
    assert "TEST_ENV_2" in env
    assert env["TEST_ENV_1"] == "1"
    assert env["TEST_ENV_2"] == "2"


@patch("runem.run_command.Popen", autospec=True, return_value=MockPopen())
def test_run_command_with_env_verbose(mock_popen: Mock) -> None:
    """Tests that the env is handled and logged out in verbose mode."""
    # capture any prints the run_command() does, should be none in verbose=False mode
    with io.StringIO() as buf, redirect_stdout(buf):
        output = runem.run_command.run_command(
            cmd=["ls"],
            label="test command",
            verbose=True,
            env_overrides={"TEST_ENV_1": "1", "TEST_ENV_2": "2"},
        )
        run_command_stdout = buf.getvalue()
    assert output == "test output"
    assert run_command_stdout == (
        "runem: running: start: test command: ls\n"
        "runem: ENV OVERRIDES: TEST_ENV_1='1' TEST_ENV_2='2' ls\n"
        "| test command: test output\n"
        "runem: running: done: test command: ls\n"
    )
    assert len(mock_popen.call_args) == 2
    assert mock_popen.call_args[0] == (["ls"],)
    call_ctx = mock_popen.call_args[1]
    env = call_ctx["env"]
    assert "TEST_ENV_1" in env
    assert "TEST_ENV_2" in env
    assert env["TEST_ENV_1"] == "1"
    assert env["TEST_ENV_2"] == "2"


@patch(
    "runem.run_command.Popen",
    autospec=True,
    return_value=MockPopen(
        returncode=1,
    ),
)
def test_run_command_with_env_on_error(mock_popen: Mock) -> None:
    """Tests that the env is passed to the subprocess and prints on error."""
    # capture any prints the run_command() does, should be none in verbose=False mode
    with io.StringIO() as buf, redirect_stdout(buf):
        with pytest.raises(runem.run_command.RunCommandBadExitCode) as err_info:
            runem.run_command.run_command(
                cmd=["ls"],
                label="test command",
                verbose=False,
                env_overrides={"TEST_ENV_1": "1", "TEST_ENV_2": "2"},
            )
        run_command_stdout = buf.getvalue()

    assert "TEST_ENV_1='1' TEST_ENV_2='2'" in str(err_info.value.stdout)

    assert "" == run_command_stdout, "expected empty output when verbosity is off"
    assert len(mock_popen.call_args) == 2
    assert mock_popen.call_args[0] == (["ls"],)
    call_ctx = mock_popen.call_args[1]
    env = call_ctx["env"]
    assert "TEST_ENV_1" in env
    assert "TEST_ENV_2" in env
    assert env["TEST_ENV_1"] == "1"
    assert env["TEST_ENV_2"] == "2"


@patch("runem.run_command.Popen", autospec=True, return_value=MockPopen())
def test_run_command_basic_call_verbose_with_cwd(mock_popen: Mock) -> None:
    """Test that we get extra output when the verbose flag is set."""
    # capture any prints the run_command() does, should be informative in verbose=True mode
    with io.StringIO() as buf, redirect_stdout(buf):
        raw_output = runem.run_command.run_command(
            cmd=["ls"],
            label="test command",
            verbose=True,
            cwd=pathlib.Path("./some/test/path"),
        )
        run_command_stdout = buf.getvalue()
    assert raw_output == "test output"

    # check the log output hasn't changed. Update as needed.
    assert run_command_stdout.split("\n") == [
        "runem: running: start: test command: ls",
        "runem: cwd: some/test/path",
        "| test command: test output",
        "runem: running: done: test command: ls",
        "",
    ]
    mock_popen.assert_called_once()



import io
import typing
from contextlib import redirect_stdout
from datetime import timedelta

from runem.report import _print_reports_by_phase, report_on_run
from runem.types.common import OrderedPhases
from runem.types.types_jobs import (
    JobReturn,
    JobRunMetadata,
    JobRunMetadatasByPhase,
    JobRunReportByPhase,
    JobTiming,
)
from tests.sanitise_reports_footer import sanitise_reports_footer


def test_report_on_run_basic_call() -> None:
    job_timing_1: JobTiming = {"job": ("job 1", timedelta(seconds=0)), "commands": []}
    job_timing_2: JobTiming = {
        "job": (
            "job label 2",
            timedelta(seconds=1000, milliseconds=1, microseconds=1),
        ),
        "commands": [
            ("sub1", timedelta(seconds=500)),
            ("sub2", timedelta(seconds=500)),
        ],
    }
    job_timing_3: JobTiming = {
        "job": (
            "another job 3",
            timedelta(seconds=2),
        ),
        "commands": [],
    }
    job_timing_4: JobTiming = {
        "job": (
            "another job 4",
            timedelta(seconds=250),
        ),
        "commands": [],
    }
    job_return: JobReturn = None  # typing.Optional[JobReturnData]
    job_run_metadata_1: JobRunMetadata = (job_timing_1, job_return)
    job_run_metadata_2: JobRunMetadata = (job_timing_2, job_return)
    job_run_metadata_3: JobRunMetadata = (job_timing_3, job_return)
    job_run_metadata_4: JobRunMetadata = (job_timing_4, job_return)
    job_run_metadatas: JobRunMetadatasByPhase = {
        "phase 1": [
            job_run_metadata_1,
            job_run_metadata_2,
            job_run_metadata_3,
            job_run_metadata_4,
        ]
    }
    with io.StringIO() as buf, redirect_stdout(buf):
        system_time_spent: timedelta
        wall_clock_time_saved: timedelta
        system_time_spent, wall_clock_time_saved = report_on_run(
            phase_run_oder=("phase 1",),
            job_run_metadatas=job_run_metadatas,
            wall_clock_for_runem_main=timedelta(
                # mimic a value that is slightly larger than the largest single
                # job in the single phase.
                seconds=1000.5
            ),
        )
        run_command_stdout = buf.getvalue()
    assert run_command_stdout.split("\n") == [
        "runem: reports:",
        "runem (total wall-clock)  [1000.500000]  ████████████████████████████████",
        (
            " └phase 1 (user-time)     [1252.001001]  "
            "░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░"
        ),
        "  ├job label 2            [1000.001001]  ████████████████████████████████",
        "  │├sub1 (+)              [ 500.000000]  ················",
        "  │└sub2 (+)              [ 500.000000]  ················",
        "  ├another job 3          [   2.000000]  ▏",
        "  └another job 4          [ 250.000000]  ████████",
        "",
    ]
    # this floating-point comparison should be problematic but its' working for
    # now... for now.
    assert wall_clock_time_saved.total_seconds() == 251.501001
    assert system_time_spent.total_seconds() == 1252.001001


def test_report_on_run_reports() -> None:
    job_return_1: JobReturn = {
        "reportUrls": [
            ("dummy report label", "/dummy/report/url"),
        ]
    }
    job_return_2: JobReturn = None  # typing.Optional[JobReturnData]
    job_timing_1: JobTiming = {
        "job": ("job label 1", timedelta(seconds=0)),
        "commands": [],
    }
    job_timing_2: JobTiming = {
        "job": (
            "job label 2",
            timedelta(seconds=1000, milliseconds=1, microseconds=1),
        ),
        "commands": [],
    }
    job_timing_3: JobTiming = {
        "job": (
            "another job 3",
            timedelta(seconds=2),
        ),
        "commands": [],
    }
    job_run_metadata_1: JobRunMetadata = (job_timing_1, job_return_1)
    job_run_metadata_2: JobRunMetadata = (job_timing_2, job_return_2)
    job_run_metadata_3: JobRunMetadata = (job_timing_3, job_return_2)
    job_run_metadatas: JobRunMetadatasByPhase = {
        "phase 1": [
            job_run_metadata_1,
            job_run_metadata_2,
            job_run_metadata_3,
        ]
    }
    with io.StringIO() as buf, redirect_stdout(buf):
        report_on_run(
            phase_run_oder=("phase 1",),
            job_run_metadatas=job_run_metadatas,
            wall_clock_for_runem_main=timedelta(0),
        )
        run_command_stdout = buf.getvalue()
    assert run_command_stdout.split("\n") == [
        "runem: reports:",
        "runem (total wall-clock)  [   0.000000]",
        (
            " └phase 1 (user-time)     [1002.001001]  "
            "░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░░"
        ),
        (
            "  ├job label 2            [1000.001001]  "
            "███████████████████████████████████████▉"
        ),
        "  └another job 3          [   2.000000]  ▏",
        "runem: report: dummy report label: /dummy/report/url",
        "",
    ]


def test_sanitise_reports_footer_removes_all_bar_chars_2() -> None:
    """Tries to test that we handle all bar-graph characters.

    This ensures we know that we are sanitising the report data nicely.
    """
    phase_run_order: typing.List[str] = []
    job_run_metadatas: JobRunMetadatasByPhase = {}
    for multiplier in [1]:  # ,2,3,4,5,6,7,8,9,10]:
        fractional_times: typing.List[JobRunMetadata] = []
        for seconds in range(15):  # 1 + 2 +3.. +14 = 105 near enough to 100%
            seconds = seconds * multiplier
            job_timing: JobTiming = {
                "job": (f"job {seconds}", timedelta(seconds=seconds)),
                "commands": [],
            }
            job_return: JobReturn = None  # typing.Optional[JobReturnData]
            job_run_metadata: JobRunMetadata = (job_timing, job_return)
            fractional_times.append(job_run_metadata)
        phase_name: str = f"phase {multiplier}"
        phase_run_order.append(phase_name)
        job_run_metadatas[phase_name] = fractional_times

    with io.StringIO() as buf, redirect_stdout(buf):
        system_time_spent: timedelta
        wall_clock_time_saved: timedelta
        system_time_spent, wall_clock_time_saved = report_on_run(
            phase_run_oder=tuple(phase_run_order),
            job_run_metadatas=job_run_metadatas,
            wall_clock_for_runem_main=timedelta(
                # mimic a value that is slightly larger than the largest single
                # job in the single phase.
                seconds=15
            ),
        )
        run_command_stdout = sanitise_reports_footer(buf.getvalue())
    assert run_command_stdout == [
        "runem: reports:",
        "runem (total wall-clock)  [ <float>]",
        " └phase 1 (user-time)     [<float>]",
        "  ├job 1                  [  <float>]",
        "  ├job 2                  [  <float>]",
        "  ├job 3                  [  <float>]",
        "  ├job 4                  [  <float>]",
        "  ├job 5                  [  <float>]",
        "  ├job 6                  [  <float>]",
        "  ├job 7                  [  <float>]",
        "  ├job 8                  [  <float>]",
        "  ├job 9                  [  <float>]",
        "  ├job 10                 [ <float>]",
        "  ├job 11                 [ <float>]",
        "  ├job 12                 [ <float>]",
        "  ├job 13                 [ <float>]",
        "  └job 14                 [ <float>]",
        "",
    ]
    # this floating-point comparison should be problematic but its' working for
    # now... for now.
    assert wall_clock_time_saved.total_seconds() == 90.0
    assert system_time_spent.total_seconds() == 105.0


def test_print_reports_by_phase() -> None:
    # Test data
    phase_run_order: OrderedPhases = ("phase1", "phase2", "phase3", "phase4")

    report_data: JobRunReportByPhase = {
        "ignored_phase": [
            ("no log", "not run should not appear in logs"),
        ],
        "phase1": [("report1", "path/to/report1"), ("report2", "path/to/report2")],
        "phase2": [("report3", "path/to/report3")],
        "phase3": [],  # Empty list to test handling of empty reports
        "phase4": [
            (),  # type: ignore
        ],  # Empty tuple to test handling of empty report data
    }

    # Call the function
    with io.StringIO() as buf, redirect_stdout(buf):
        _print_reports_by_phase(phase_run_order, report_data)
        run_command_stdout = buf.getvalue()

    # Assert that log is called with the expected messages
    expected_logs = [
        "runem: report: report1: path/to/report1",
        "runem: report: report2: path/to/report2",
        "runem: report: report3: path/to/report3",
        "",
    ]
    assert run_command_stdout.split("\n") == expected_logs



import os
import pathlib
import sys
import typing

import pytest

from runem.blocking_print import _reset_console, _reset_console_for_tests

FixtureRequest = typing.Any


@pytest.fixture(autouse=True)
def use_test_rich_print(request: FixtureRequest) -> typing.Generator[None, None, None]:
    """Each test should use the test-version of the `rich` print function.

    This is so we get deterministic output with out timestamps nor auto-wrapping console
    output.
    """
    _reset_console_for_tests()
    yield
    _reset_console()


# each test runs on cwd to its temp dir
@pytest.fixture(autouse=True)
def go_to_tmp_path(request: FixtureRequest) -> typing.Generator[None, None, None]:
    # Get the fixture dynamically by its name.
    tmp_path: pathlib.Path = request.getfixturevalue("tmp_path")
    # ensure local test created packages can be imported
    sys.path.insert(0, str(tmp_path))
    # Chdir only for the duration of the test.
    origin = pathlib.Path().absolute()
    try:
        os.chdir(tmp_path)
        yield
    finally:
        os.chdir(origin)



import time
import typing

from rich.console import Console


def _reset_console() -> Console:
    """Sets the nice user-facing console.

    This function exists so we can reset the output from tests.
    """
    global RICH_CONSOLE  # pylint: disable=global-statement

    RICH_CONSOLE = Console(
        log_path=False,  # Do NOT print the source path.
        # We allow markup here, BUT stdout/stderr from other procs should have
        # `escape()` called on them so they don't error here.
        # This means 'rich' effects/colors can be judiciously applied:
        # e.g. `[blink]Don't Panic![/blink]`.
        markup=True,
        # `highlight` is what colourises string and number in print() calls.
        # We do not want this to be auto-magic.
        highlight=False,
        # `soft_wrap=True` disables word-wrap & cropping by default:
        #  - `soft_wrap` reads like a misnomer to me
        soft_wrap=True,
    )
    return RICH_CONSOLE


# Overridden in tests.
RICH_CONSOLE: Console = _reset_console()


def _reset_console_for_tests() -> None:
    """Overrides the console output with more deterministic version.

    ... so we can have deterministic tests.
    """
    global RICH_CONSOLE  # pylint: disable=global-statement
    RICH_CONSOLE = Console(
        log_path=False,  # Do NOT print the source path.
        log_time=False,  # Do not prefix with log time e.g. `[time] log message`.
        markup=True,  # Allow some markup e.g. `[blink]Don't Panic![/blink]`.
        highlight=False,
        width=999,  # A very wide width.
    )


def blocking_print(
    msg: str = "",
    end: typing.Optional[str] = None,
    max_retries: int = 5,
    sleep_time_s: float = 0.1,
) -> None:
    """Attempt to print a message, retrying on BlockingIOError.

    Sometimes in long-lasting jobs, that produce lots of output, we hit
    BlockingIOError where we can't print to screen because the buffer is full or
    already being written to (for example), i.e. the `print` would need to be a
    'blocking' call, which it is not.
    """
    if end is None:
        end = "\n"
    for _ in range(max_retries):
        try:
            RICH_CONSOLE.print(msg, end=end)
            break  # Success, exit the retry loop
        except BlockingIOError:
            time.sleep(sleep_time_s)  # Wait a bit for the buffer to clear
    else:
        # Optional: handle the failure to print after all retries
        pass



import typing

# Define type variables for key and value to be used in the custom dictionary
K = typing.TypeVar("K")
V = typing.TypeVar("V")


class InformativeDict(typing.Dict[K, V], typing.Generic[K, V]):
    """A dictionary type that prints out the available keys."""

    def __getitem__(self, key: K) -> V:
        """Attempt to get item, raising a detailed exception if the key is not found."""
        try:
            return super().__getitem__(key)
        except KeyError:
            available_keys: typing.Iterable[str] = (str(k) for k in self.keys())
            raise KeyError(
                f"Key '{key}' not found. Available keys: {', '.join(available_keys)}"
            ) from None


class ReadOnlyInformativeDict(InformativeDict[K, V], typing.Generic[K, V]):
    """A read-only variant of the above."""

    def __setitem__(self, key: K, value: V) -> None:
        """Readonly object, setitem disallowed."""
        raise NotImplementedError("This dictionary is read-only")

    def __delitem__(self, key: K) -> None:
        """Readonly object, delitem disallowed."""
        raise NotImplementedError("This dictionary is read-only")

    def pop(self, *args: typing.Any, **kwargs: typing.Any) -> V:
        """Readonly object, pop disallowed."""
        raise NotImplementedError("This dictionary is read-only")

    def popitem(self) -> typing.Tuple[K, V]:
        """Readonly object, popitem disallowed."""
        raise NotImplementedError("This dictionary is read-only")

    def clear(self) -> None:
        """Readonly object, clear disallowed."""
        raise NotImplementedError("This dictionary is read-only")

    def update(self, *args: typing.Any, **kwargs: typing.Any) -> None:
        """Readonly object, update disallowed."""
        raise NotImplementedError("This dictionary is read-only")



import pathlib
from argparse import Namespace
from collections import defaultdict
from typing import Dict
from unittest.mock import MagicMock

import pytest

from runem.command_line import initialise_options
from runem.config_metadata import ConfigMetadata
from runem.types.runem_config import JobConfig, PhaseGroupedJobs

Options = Dict[str, bool]


@pytest.fixture(name="config_metadata")
def config_metadata_fixture() -> ConfigMetadata:
    config_file_path = pathlib.Path(__file__).parent / ".runem.yml"
    expected_job: JobConfig = {
        "addr": {
            "file": "test_config_parse.py",
            "function": "test_parse_config",
        },
        "label": "dummy job label",
        "when": {
            "phase": "dummy phase 1",
            "tags": {"dummy tag 1", "dummy tag 2"},
        },
    }
    expected_jobs: PhaseGroupedJobs = defaultdict(list)
    expected_jobs["dummy phase 1"] = [
        expected_job,
    ]
    return ConfigMetadata(
        cfg_filepath=config_file_path,
        phases=("dummy phase 1",),
        options_config=(
            {"name": "option1", "default": True},
            {"name": "option2", "default": False},
            {"name": "option3", "default": True},
            {"name": "option4", "default": False},
        ),
        file_filters={
            # "dummy tag": {
            #     "tag": "dummy tag",
            #     "regex": ".*1.txt",  # should match just one file
            # }
        },
        hook_manager=MagicMock(),
        jobs=expected_jobs,
        all_job_names=set(("dummy job label",)),
        all_job_phases=set(("dummy phase 1",)),
        all_job_tags=set(
            (
                "dummy tag 2",
                "dummy tag 1",
            )
        ),
    )


def test_initialise_options_no_overrides(config_metadata: ConfigMetadata) -> None:
    args = Namespace(overrides_on=[], overrides_off=[])
    options = initialise_options(config_metadata, args)
    assert options == {
        "option1": True,
        "option2": False,
        "option3": True,
        "option4": False,
    }


def test_initialise_options_overrides_on(config_metadata: ConfigMetadata) -> None:
    args = Namespace(overrides_on=["option2", "option4"], overrides_off=[])
    options = initialise_options(config_metadata, args)
    assert options == {
        "option1": True,
        "option2": True,
        "option3": True,
        "option4": True,
    }


def test_initialise_options_overrides_off(config_metadata: ConfigMetadata) -> None:
    args = Namespace(overrides_on=[], overrides_off=["option1", "option3"])
    options = initialise_options(config_metadata, args)
    assert options == {
        "option1": False,
        "option2": False,
        "option3": False,
        "option4": False,
    }


def test_initialise_options_overrides_on_and_off(
    config_metadata: ConfigMetadata,
) -> None:
    args = Namespace(overrides_on=["option2"], overrides_off=["option1"])
    options = initialise_options(config_metadata, args)
    assert options == {
        "option1": False,
        "option2": True,
        "option3": True,
        "option4": False,
    }



import typing

from runem.blocking_print import blocking_print


def log(
    msg: str = "",
    prefix: typing.Optional[bool] = None,
    end: typing.Optional[str] = None,
) -> None:
    """Thin wrapper around 'print', change the 'msg' & handles system-errors.

    One way we change it is to decorate the output with 'runem'

    Parameters:
        msg: str - the message to log out. Any `rich` markup will be escaped
                   and not applied.
        decorate: str - whether to add runem-specific prefix text. We do this
                        to identify text that comes from the app vs text that
                        comes from hooks or other third-parties.
        end: Optional[str] - same as the end option used by `print()` and
                             `rich`
    Returns:
        Nothing
    """
    # Remove any markup as it will probably error, if unsanitised.
    # msg = escape(msg)

    if prefix is None:
        prefix = True

    if prefix:
        # Make it clear that the message comes from `runem` internals.
        msg = f"[light_slate_grey]runem[/light_slate_grey]: {msg}"

    # print in a blocking manner, waiting for system resources to free up if a
    # runem job is contending on stdout or similar.
    blocking_print(msg, end=end)


def warn(msg: str) -> None:
    log(f"[yellow]WARNING[/yellow]: {msg}")


def error(msg: str) -> None:
    log(f"[red]ERROR[/red]: {msg}")



import io
import os
import pathlib
import typing
from contextlib import redirect_stdout
from unittest.mock import Mock, patch

import pytest

from runem.config import _find_local_configs, load_project_config, load_user_configs
from runem.types.runem_config import Config, GlobalConfig


def test_load_project_config(tmp_path: pathlib.Path) -> None:
    config_gen_path: pathlib.Path = tmp_path / ".runem.yml"
    config_gen_path.write_text(
        "- config:\n"
        "    phases:\n"
        "      - mock phase\n"
        "    min_version: 0.0.0\n"
        "    files:\n"
        "    options:\n"
    )

    # set the working dir to where the config is
    os.chdir(tmp_path)

    loaded_config: Config
    config_read_path: pathlib.Path
    loaded_config, config_read_path = load_project_config()
    expected_config: Config = [
        {
            "config": {
                "files": None,
                "options": None,
                "phases": ("mock phase",),
                "min_version": "0.0.0",
            }
        }
    ]
    assert loaded_config == expected_config
    assert config_read_path == config_gen_path


def test_load_project_config_with_no_phases(tmp_path: pathlib.Path) -> None:
    config_gen_path: pathlib.Path = tmp_path / ".runem.yml"
    config_gen_path.write_text(
        (
            "- config:\n"  # ln 1
            "    files:\n"  # ln 2
            "    options:\n"  # ln 3
        )
    )

    # set the working dir to where the config is
    os.chdir(tmp_path)

    loaded_config: Config
    config_read_path: pathlib.Path
    loaded_config, config_read_path = load_project_config()
    expected_config: Config = [
        {
            "config": {  # type: ignore # intentionally testing for missing 'phases'
                "files": None,
                "options": None,
            }
        }
    ]
    assert loaded_config == expected_config
    assert config_read_path == config_gen_path


def test_load_project_config_with_min_version(tmp_path: pathlib.Path) -> None:
    config_gen_path: pathlib.Path = tmp_path / ".runem.yml"
    config_gen_path.write_text(
        (
            "- config:\n"  # ln 1
            "    files:\n"  # ln 2
            "    options:\n"  # ln 3
            "    min_version: 9999.99999.9999"  # a large min-version
        )
    )

    # set the working dir to where the config is
    os.chdir(tmp_path)

    with io.StringIO() as buf, redirect_stdout(buf):
        with pytest.raises(SystemExit):
            load_project_config()
        runem_stdout: str = buf.getvalue()
        assert (
            "runem: .runem.yml config requires runem '9999.99999.9999', you have"
            in runem_stdout
        )


def test_load_project_config_with_global_last(tmp_path: pathlib.Path) -> None:
    config_gen_path: pathlib.Path = tmp_path / ".runem.yml"
    config_gen_path.write_text(
        (
            "- job:\n"  # some job
            "    addr:\n"
            "      file: scripts/test_hooks/py.py\n"
            "      function: _job_py_pytest\n"
            "    label: pytest\n"
            "    when:\n"
            "      phase: analysis\n"
            "      tags:\n"
            "        - py\n"
            "        - unit test\n"
            "        - test\n"
            "- config:\n"  # the global config last
            "    files:\n"
            "    options:\n"
        )
    )

    # set the working dir to where the config is
    os.chdir(tmp_path)

    loaded_config: Config
    config_read_path: pathlib.Path
    loaded_config, config_read_path = load_project_config()
    global_config: GlobalConfig = {  # type: ignore # intentionally testing for missing 'phases'
        "files": None,
        "options": None,
    }
    expected_config: Config = [
        {
            "job": {
                "addr": {
                    "file": "scripts/test_hooks/py.py",
                    "function": "_job_py_pytest",
                },
                "label": "pytest",
                "when": {
                    "phase": "analysis",
                    "tags": ["py", "unit test", "test"],  # type: ignore
                },
            }
        },
        {
            "config": global_config,
        },
    ]
    assert loaded_config == expected_config
    assert config_read_path == config_gen_path


@patch(
    "runem.config._find_config_file",
    return_value=(pathlib.Path("dummy path"), None),
)
@patch("pathlib.Path.exists", return_value=True)
def test_find_local_configs(
    path_exists_mock: Mock,
    find_config_file_mock: Mock,
) -> None:
    configs: typing.List[pathlib.Path] = _find_local_configs()
    assert len(configs) == 3
    assert configs == [
        pathlib.Path("dummy path"),
        pathlib.Path("dummy path"),
        pathlib.Path("~/.runem.user.yml"),
    ]
    find_config_file_mock.assert_called()
    path_exists_mock.assert_called()


@patch(
    "runem.config._find_config_file",
    return_value=(None, None),
)
@patch("pathlib.Path.exists", return_value=False)
def test_find_local_configs_finds_nothing(
    path_exists_mock: Mock,
    find_config_file_mock: Mock,
) -> None:
    configs: typing.List[pathlib.Path] = _find_local_configs()
    assert len(configs) == 0
    assert not configs
    find_config_file_mock.assert_called()
    path_exists_mock.assert_called()


@patch(
    "runem.config.load_and_parse_config",
    return_value="DUMMY CONFIG",
)
@patch(
    "runem.config._find_config_file",
    return_value=(pathlib.Path("dummy path"), None),
)
@patch("pathlib.Path.exists", return_value=True)
def test_load_user_configs(
    path_exists_mock: Mock,
    find_config_file_mock: Mock,
    load_and_parse_config_mock: Mock,
) -> None:
    configs: typing.List[typing.Tuple[Config, pathlib.Path]] = load_user_configs()
    assert len(configs) == 3
    expected_config: typing.List[typing.Tuple[Config, pathlib.Path]] = [
        ("DUMMY CONFIG", pathlib.Path("dummy path")),  # type:ignore
        ("DUMMY CONFIG", pathlib.Path("dummy path")),  # type:ignore
        ("DUMMY CONFIG", pathlib.Path("~/.runem.user.yml")),  # type:ignore
    ]
    assert configs == expected_config
    find_config_file_mock.assert_called()
    path_exists_mock.assert_called()
    load_and_parse_config_mock.assert_called()



import pathlib
from argparse import Namespace
from collections import defaultdict
from typing import List, Optional
from unittest.mock import MagicMock, Mock, patch

import pytest

from runem.config_metadata import ConfigMetadata
from runem.files import find_files
from runem.informative_dict import InformativeDict
from runem.types.filters import FilePathListLookup


def _prep_config(
    tmp_path: pathlib.Path,
    check_modified_files: bool,
    check_head_files: bool,
    always_files: Optional[List[str]],
    git_since_branch: Optional[str],
) -> ConfigMetadata:
    config_metadata: ConfigMetadata = ConfigMetadata(
        cfg_filepath=tmp_path / ".runem.yml.no-exist",  # only the path is used
        phases=("dummy phase 1",),
        options_config=tuple(),
        file_filters={
            "dummy tag": {
                "tag": "dummy tag",
                "regex": ".*1.txt",  # should match just one file
            }
        },
        hook_manager=MagicMock(),
        jobs=defaultdict(list),
        all_job_names=set(),
        all_job_phases=set(),
        all_job_tags=set(),
    )
    config_metadata.set_cli_data(
        args=Namespace(
            check_modified_files=check_modified_files,
            check_head_files=check_head_files,
            always_files=always_files,
            git_since_branch=git_since_branch,
        ),
        jobs_to_run=set(),  # JobNames,
        phases_to_run=set(),  # JobPhases,
        tags_to_run=set(),  # ignored JobTags,
        tags_to_avoid=set(),  # ignored  JobTags,
        options=InformativeDict({}),  # Options,
    )

    return config_metadata


@pytest.mark.parametrize(
    "check_head_files",
    [
        True,
        False,
    ],
)
@pytest.mark.parametrize(
    "check_modified_files",
    [
        True,
        False,
    ],
)
@pytest.mark.parametrize(
    "always_files",
    [
        ["1.always", "2.no_exist.always", "3.always"],
        [],  # empty array
        None,  # trigger defaults
    ],
)
@patch(
    "runem.files.subprocess_check_output",
)
def test_find_files_basic(
    mock_subprocess_check_output: Mock,
    always_files: Optional[List[str]],
    check_modified_files: bool,
    check_head_files: bool,
    tmp_path: pathlib.Path,
) -> None:
    file_strings: List[str] = []
    for file_str in ("test_file_1.txt", "test_file_2.txt"):
        test_file: pathlib.Path = tmp_path / file_str
        test_file.touch()  # write some empty string aka 'touch' the file
        file_strings.append(str(test_file))
    mock_subprocess_check_output.return_value = str.encode("\n".join(file_strings))

    created_always_files: Optional[List[str]] = None
    if always_files is not None:
        created_always_files = []
        for always_file in always_files:
            file_path: pathlib.Path = tmp_path / always_file
            if "no_exist" not in always_file:
                # only create files if they are NOT tagged with 'no_exist'
                file_path.touch()
            created_always_files.append(str(file_path))

    config_metadata = _prep_config(
        tmp_path,
        check_modified_files=check_modified_files,
        check_head_files=check_head_files,
        always_files=created_always_files,
        git_since_branch=None,
    )
    results: FilePathListLookup = find_files(config_metadata)
    if check_modified_files and check_head_files:
        assert mock_subprocess_check_output.call_count == 3, (
            "twice for modified, once for head"
        )
        assert results == {
            "dummy tag": [file_strings[0]]  # we filter in only the *1* files.
        }
    elif check_modified_files:
        assert mock_subprocess_check_output.call_count == 2, "twice for modified"
        assert results == {
            "dummy tag": [file_strings[0]]  # we filter in only the *1* files.
        }
    else:
        assert mock_subprocess_check_output.call_count == 1, "once for git ls-files"
        assert results == {
            "dummy tag": [file_strings[0]]  # we filter in only the *1* files.
        }


@pytest.mark.parametrize(
    "git_since_branch",
    [
        None,
        "/dummy/branch/name",
    ],
)
@patch(
    "runem.files.subprocess_check_output",
)
def test_find_files_git_since_branch(
    mock_subprocess_check_output: Mock,
    git_since_branch: Optional[str],
    tmp_path: pathlib.Path,
) -> None:
    file_strings: List[str] = []
    for file_str in ("test_file_1.txt", "test_file_2.txt"):
        test_file: pathlib.Path = tmp_path / file_str
        test_file.touch()  # write some empty string aka 'touch' the file
        file_strings.append(str(test_file))
    mock_subprocess_check_output.return_value = str.encode("\n".join(file_strings))

    config_metadata = _prep_config(
        tmp_path,
        check_modified_files=False,
        check_head_files=False,
        always_files=None,
        git_since_branch=git_since_branch,
    )
    results: FilePathListLookup = find_files(config_metadata)
    if git_since_branch is None:
        assert mock_subprocess_check_output.call_count == 1, "once for git ls-files"
        assert results == {
            "dummy tag": [file_strings[0]]  # we filter in only the *1* files.
        }
    else:
        assert mock_subprocess_check_output.call_count == 1, "once for git ls-files"
        assert results == {
            "dummy tag": [file_strings[0]]  # we filter in only the *1* files.
        }



import copy
import pathlib
import sys
import typing
from collections import defaultdict
from collections.abc import Iterable

from runem.config_metadata import ConfigMetadata
from runem.hook_manager import HookManager
from runem.job import Job
from runem.job_wrapper import get_job_wrapper
from runem.log import error, log, warn
from runem.types.common import JobNames, JobPhases, JobTags, OrderedPhases, PhaseName
from runem.types.errors import FunctionNotFound, SystemExitBad
from runem.types.filters import TagFileFilter, TagFileFilters
from runem.types.hooks import HookName
from runem.types.runem_config import (
    Config,
    ConfigNodes,
    GlobalConfig,
    GlobalSerialisedConfig,
    HookConfig,
    Hooks,
    HookSerialisedConfig,
    JobConfig,
    JobSerialisedConfig,
    JobWhen,
    OptionConfigs,
    PhaseGroupedJobs,
    TagFileFilterSerialised,
)


def _parse_global_config(
    global_config: GlobalConfig,
) -> typing.Tuple[OrderedPhases, OptionConfigs, TagFileFilters]:
    """Parses and validates a global-config entry read in from disk.

    Returns the phases in the order we want to run them
    """
    options: OptionConfigs = ()
    if "options" in global_config and global_config["options"]:
        options = tuple(
            option_serialised["option"]
            for option_serialised in global_config["options"]
        )

    file_filters: TagFileFilters = {}
    if "files" in global_config and global_config["files"]:
        file_filter: TagFileFilterSerialised
        serialised_filters: typing.List[TagFileFilterSerialised] = global_config[
            "files"
        ]
        for file_filter in serialised_filters:
            actual_filter: TagFileFilter = file_filter["filter"]
            tag = actual_filter["tag"]
            file_filters[tag] = actual_filter

    phases: OrderedPhases = tuple()
    if "phases" in global_config:
        phases = global_config["phases"]
    return phases, options, file_filters


def parse_hook_config(
    hook: HookConfig,
    cfg_filepath: pathlib.Path,
) -> None:
    """Get the hook information, verifying validity."""
    try:
        if not HookManager.is_valid_hook_name(hook["hook_name"]):
            raise ValueError(
                f"invalid hook-name '{str(hook['hook_name'])}'. "
                f"Valid hook names are: {[hook.value for hook in HookName]}"
            )
        # cast the hook-name to a HookName type
        hook["hook_name"] = HookName(hook["hook_name"])
        get_job_wrapper(hook, cfg_filepath)
    except KeyError as err:
        raise ValueError(
            f"hook config entry is missing '{err.args[0]}' key. Have {tuple(hook.keys())}"
        ) from err
    except FunctionNotFound as err:
        error(f"Whilst loading hook '{str(hook['hook_name'])}'. {str(err)}")
        raise SystemExitBad(2) from err


def _parse_job(  # noqa: C901
    cfg_filepath: pathlib.Path,
    job: JobConfig,
    in_out_tags: JobTags,
    in_out_jobs_by_phase: PhaseGroupedJobs,
    in_out_job_names: JobNames,
    in_out_phases: JobPhases,
    phase_order: OrderedPhases,
    warn_missing_phase: bool = True,
) -> None:
    """Parse an individual job."""
    job_name: str = Job.get_job_name(job)
    job_names_used = job_name in in_out_job_names
    if job_names_used:
        error(
            "duplicate job label!"
            f"\t'{job['label']}' is used twice or more in {str(cfg_filepath)}"
        )
        sys.exit(1)

    try:
        # try and load the function _before_ we schedule it's execution
        get_job_wrapper(job, cfg_filepath)
    except FunctionNotFound as err:
        error(f"Whilst loading job '{job['label']}'. {str(err)}")
        raise SystemExitBad(2) from err

    try:
        phase_id: PhaseName = job["when"]["phase"]
    except KeyError:
        try:
            fallback_phase = phase_order[0]
            if warn_missing_phase:
                warn(f"no phase found for '{job_name}', using '{fallback_phase}'")
        except IndexError:
            fallback_phase = "<NO PHASES FOUND>"
            if warn_missing_phase:
                warn(
                    (
                        f"no phases found for '{job_name}', "
                        f"or in '{str(cfg_filepath)}', "
                        f"using '{fallback_phase}'"
                    )
                )
        phase_id = fallback_phase
    in_out_jobs_by_phase[phase_id].append(job)

    in_out_job_names.add(job_name)
    in_out_phases.add(phase_id)
    job_tags: typing.Optional[JobTags] = Job.get_job_tags(job)
    if job_tags:
        in_out_tags.update(job_tags)


def parse_job_config(
    cfg_filepath: pathlib.Path,
    job: JobConfig,
    in_out_tags: JobTags,
    in_out_jobs_by_phase: PhaseGroupedJobs,
    in_out_job_names: JobNames,
    in_out_phases: JobPhases,
    phase_order: OrderedPhases,
    silent: bool = False,
) -> None:
    """Parses and validates a job-entry read in from disk.

    Returns the tags generated
    """
    try:
        # if there is more than one cwd, duplicate the job for each cwd
        generated_jobs: typing.List[JobConfig] = []
        have_ctw_cwd: bool = (("ctx" in job) and (job["ctx"] is not None)) and (
            ("cwd" in job["ctx"]) and (job["ctx"]["cwd"] is not None)
        )
        if (not have_ctw_cwd) or isinstance(
            job["ctx"]["cwd"],  # type: ignore # handled above
            str,
        ):
            # if
            # - we don't have a cwd, ctx
            # - or if the cwd is just a string, it's a path, just use it
            generated_jobs.append(job)
        else:
            assert job["ctx"] is not None
            assert job["ctx"]["cwd"] is not None
            assert isinstance(job["ctx"]["cwd"], Iterable)
            assert isinstance(job["ctx"]["cwd"], list)
            cwd_list: typing.List[str] = job["ctx"]["cwd"]
            cwd: str
            for cwd in cwd_list:
                specialised_job_for_cwd = copy.deepcopy(job)
                # overwrite the list of cwd paths with just the single instance
                assert (
                    "ctx" in specialised_job_for_cwd and specialised_job_for_cwd["ctx"]
                ), specialised_job_for_cwd
                assert (
                    "cwd" in specialised_job_for_cwd["ctx"]
                    and specialised_job_for_cwd["ctx"]["cwd"]
                ), specialised_job_for_cwd["ctx"].keys()
                specialised_job_for_cwd["ctx"]["cwd"] = cwd

                # add the last directory name from the 'cwd' path as a tag for
                # easy reference to the job-task by its path
                when: JobWhen = specialised_job_for_cwd.get("when", {})
                when["tags"] = set(when.get("tags", set()))
                cwd_path: pathlib.Path = pathlib.Path(cwd)
                when["tags"].add(cwd_path.name)
                specialised_job_for_cwd["when"] = when
                specialised_job_for_cwd["ctx"]["cwd"] = cwd

                # update the label to reflect the specialisation
                specialised_job_for_cwd["label"] = f"{job['label']}({cwd})"
                generated_jobs.append(specialised_job_for_cwd)

        for generated_job in generated_jobs:
            _parse_job(
                cfg_filepath,
                generated_job,
                in_out_tags,
                in_out_jobs_by_phase,
                in_out_job_names,
                in_out_phases,
                phase_order,
                warn_missing_phase=(not silent),
            )
    except KeyError as err:
        raise ValueError(
            f"job config entry is missing '{err.args[0]}' data. Have {job}"
        ) from err


def parse_config(  # noqa: C901
    config: Config,
    cfg_filepath: pathlib.Path,
    silent: bool = False,
    hooks_only: bool = False,
) -> typing.Tuple[
    Hooks,  # hooks:
    OrderedPhases,  # phase_order:
    OptionConfigs,  # options:
    TagFileFilters,  # file_filters:
    PhaseGroupedJobs,  # jobs_by_phase:
    JobNames,  # job_names:
    JobPhases,  # job_phases:
    JobTags,  # tags:
]:
    """Validates and restructure the config to make it more convenient to use."""
    jobs_by_phase: PhaseGroupedJobs = defaultdict(list)
    job_names: JobNames = set()
    job_phases: JobPhases = set()
    tags: JobTags = set()
    entry: ConfigNodes
    seen_global: bool = False
    phase_order: OrderedPhases = ()
    options: OptionConfigs = ()
    file_filters: TagFileFilters = {}
    hooks: Hooks = defaultdict(list)

    # first search for the global config
    for entry in config:
        # we apply a type-ignore here as we know (for now) that jobs have "job"
        # keys and global configs have "global" keys
        isinstance_job: bool = "job" in entry
        if isinstance_job:
            continue

        # we apply a type-ignore here as we know (for now) that jobs have "job"
        # keys and global configs have "global" keys
        isinstance_global: bool = "config" in entry
        if isinstance_global:
            if seen_global:
                raise ValueError(
                    "Found two global config entries, expected only one 'config' section. "
                    f"second one is {entry}"
                )
            seen_global = True
            global_entry: GlobalSerialisedConfig = entry  # type: ignore  # see above
            global_config: GlobalConfig = global_entry["config"]
            phase_order, options, file_filters = _parse_global_config(global_config)
            continue

        # we apply a type-ignore here as we know (for now) that jobs have "job"
        # keys and global configs have "global" keys
        isinstance_hooks: bool = "hook" in entry
        if isinstance_hooks:
            hook_entry: HookSerialisedConfig = entry  # type: ignore  # see above
            hook: HookConfig = hook_entry["hook"]
            parse_hook_config(hook, cfg_filepath)

            # if we get here we have validated the hook, add it to the hooks list
            hook_name: HookName = hook["hook_name"]
            hooks[hook_name].append(hook)

            # continue to the next element and do NOT error
            continue

        # not a global or a job entry, what is it
        raise RuntimeError(f"invalid 'job', 'hook, or 'global' config entry, {entry}")

    if not phase_order:
        if not hooks_only:
            if silent:  # pragma: no cover
                pass
            else:
                warn("phase ordering not configured! Runs will be non-deterministic!")
            phase_order = tuple(job_phases)

    # now parse out the job_configs
    for entry in config:
        isinstance_job_2: bool = "job" in entry
        if not isinstance_job_2:
            continue

        job_entry: JobSerialisedConfig = entry  # type: ignore  # see above
        job: JobConfig = job_entry["job"]
        parse_job_config(
            cfg_filepath,
            job,
            in_out_tags=tags,
            in_out_jobs_by_phase=jobs_by_phase,
            in_out_job_names=job_names,
            in_out_phases=job_phases,
            phase_order=phase_order,
            silent=silent,
        )
    return (
        hooks,
        phase_order,
        options,
        file_filters,
        jobs_by_phase,
        job_names,
        job_phases,
        tags,
    )


def generate_config(
    cfg_filepath: pathlib.Path,
    hooks: Hooks,
    phase_order: OrderedPhases,
    verbose: bool,
    options: OptionConfigs,
    file_filters: TagFileFilters,
    jobs_by_phase: PhaseGroupedJobs,
    job_names: JobNames,
    job_phases: JobPhases,
    tags: JobTags,
) -> ConfigMetadata:
    """Constructs the ConfigMetadata from parsed config parts."""
    return ConfigMetadata(
        cfg_filepath,
        phase_order,
        options,
        file_filters,
        HookManager(hooks, verbose),
        jobs_by_phase,
        job_names,
        job_phases,
        tags,
    )


def _load_user_hooks_from_config(
    user_config: Config,
    cfg_filepath: pathlib.Path,
    silent: bool,
) -> Hooks:
    hooks: Hooks
    (
        hooks,
        _,
        _,
        _,
        _,
        _,
        _,
        _,
    ) = parse_config(user_config, cfg_filepath, silent, hooks_only=True)
    return hooks


def load_config_metadata(
    config: Config,
    cfg_filepath: pathlib.Path,
    user_configs: typing.List[typing.Tuple[Config, pathlib.Path]],
    silent: bool = False,
    verbose: bool = False,
) -> ConfigMetadata:
    hooks: Hooks
    phase_order: OrderedPhases
    options: OptionConfigs
    file_filters: TagFileFilters
    jobs_by_phase: PhaseGroupedJobs
    job_names: JobNames
    job_phases: JobPhases
    tags: JobTags
    (
        hooks,
        phase_order,
        options,
        file_filters,
        jobs_by_phase,
        job_names,
        job_phases,
        tags,
    ) = parse_config(config, cfg_filepath, silent)

    user_config: Config
    user_config_path: pathlib.Path
    for user_config, user_config_path in user_configs:
        user_hooks: Hooks = _load_user_hooks_from_config(
            user_config, user_config_path, silent
        )
        if user_hooks:
            if verbose:
                log(f"hooks: loading user hooks from '{str(user_config_path)}'")
        hook_name: HookName
        hooks_for_name: typing.List[HookConfig]
        for hook_name, hooks_for_name in user_hooks.items():
            hooks[hook_name].extend(hooks_for_name)
            if verbose:
                log(
                    f"hooks:\tadded {len(hooks_for_name)} user hooks for '{str(hook_name)}'"
                )

    return generate_config(
        cfg_filepath,
        hooks,
        phase_order,
        verbose,
        options,
        file_filters,
        jobs_by_phase,
        job_names,
        job_phases,
        tags,
    )



import re
import typing
from collections import defaultdict
from pathlib import Path
from subprocess import check_output as subprocess_check_output

from runem.config_metadata import ConfigMetadata
from runem.types.filters import FilePathListLookup


def find_files(config_metadata: ConfigMetadata) -> FilePathListLookup:
    """Ronseal function, finds files.

    For brevity we use git-ls-files, which, for now, limits us to working in git projects.

    Previous incarnations used various methods:
        - in bash we used find with hard-coded exclude filters
        - the bash version was ported to python using os.walk()
        - the limitations of using hardcoded values were overcome using
          `gitignore-parser` but it is too slow on larger projects and 'runem' becomes
          the largest overhead, which isn't acceptable

    TODO: make this function support plugins.
    """
    file_lists: FilePathListLookup = defaultdict(list)

    file_paths: typing.List[str] = []

    if (
        config_metadata.args.check_modified_files
        or config_metadata.args.check_head_files
        or (config_metadata.args.git_since_branch is not None)
    ):
        if config_metadata.args.check_modified_files:
            # get modified, un-staged files first
            file_paths.extend(
                subprocess_check_output(
                    "git diff --name-only",
                    shell=True,
                )
                .decode("utf-8")
                .splitlines()
            )
            # now get modified, staged files first
            file_paths.extend(
                subprocess_check_output(
                    "git diff --name-only --staged",
                    shell=True,
                )
                .decode("utf-8")
                .splitlines()
            )

        if config_metadata.args.check_head_files:
            # Fetching modified and added files from the HEAD commit
            file_paths.extend(
                subprocess_check_output(
                    "git diff-tree --no-commit-id --name-only -r HEAD",
                    shell=True,
                )
                .decode("utf-8")
                .splitlines()
            )

        if config_metadata.args.git_since_branch is not None:
            # Add all files changed since a particular branch e..g `origin/main`
            # Useful for quickly checking branches before pushing.
            # NOTE: without dependency checking this might report false-positives.
            target_branch: str = config_metadata.args.git_since_branch
            file_paths.extend(
                subprocess_check_output(
                    f"git diff --name-only {target_branch}...HEAD",
                    shell=True,
                )
                .decode("utf-8")
                .splitlines()
            )
        # ensure files are unique, and still on disk i.e. filter-out deleted files
        file_paths = list(
            {file_path for file_path in file_paths if Path(file_path).exists()}
        )

    else:
        # fall-back to all files
        file_paths = (
            subprocess_check_output(
                "git ls-files",
                shell=True,
            )
            .decode("utf-8")
            .splitlines()
        )

    if config_metadata.args.always_files is not None:
        # a poor-man's version of adding path-regex's
        existent_files = [
            filepath
            for filepath in config_metadata.args.always_files
            if Path(filepath).exists()
        ]
        file_paths.extend(existent_files)

    _bucket_file_by_tag(
        file_paths,
        config_metadata,
        in_out_file_lists=file_lists,
    )

    # now ensure the file lists are sorted so we get deterministic behaviour in tests
    for job_type in file_lists:
        file_lists[job_type] = sorted(file_lists[job_type])
    return file_lists


def _bucket_file_by_tag(  # noqa: C901 # pylint: disable=too-many-branches
    file_paths: typing.List[str],
    config_metadata: ConfigMetadata,
    in_out_file_lists: FilePathListLookup,
) -> None:
    """Groups files by the file.filters iin the config."""
    for file_path in file_paths:
        for tag, file_filter in config_metadata.file_filters.items():
            if re.search(file_filter["regex"], file_path):
                in_out_file_lists[tag].append(file_path)



from runem.run_command import RunemJobError


class IntentionalTestError(RunemJobError):
    pass

    def __init__(self) -> None:
        super().__init__(friendly_message="expected test error", stdout="dummy stdout")
